/*
 * Copyright 2013 <James.Bottomley@HansenPartnership.com>
 *
 * see COPYING file
 */

#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <limits.h>

#include <openssl/x509.h>

#define __STDC_VERSION__ 199901L
#include <efi.h>

#include <kernel_efivars.h>
#include <guid.h>
#include <sha256.h>
#include <version.h>
#include "efiauthenticated.h"

#define ARRAY_SIZE(a) (sizeof (a) / sizeof ((a)[0]))

static void
usage(const char *progname)
{
	printf("Usage: %s: [-v <var>] [-s <list>[-<entry>]] [-o <file>] [-d <dir>]\n", progname);
}

static void
help(const char *progname)
{
	usage(progname);
	printf("List the contents of the UEFI signature databases\n\n"
	       "Options:\n"
	       "\t-v <var>\tlist only the contents of <var>\n"
	       "\t-s <list>[-<entry>]\tlist only a given signature list (and optionally\n"
	       "\t\tonly a given entry in that list\n"
	       "\t-o <file>\toutput the requested signature lists to <file>\n"
	       "\t-d <dir>\toutput dir for signature lists\n"
	       );
}

void
parse_db(const char *name, uint8_t *data, uint32_t len, int sig, int entry,
    const char *dir)
{
	char *p = NULL, pbuf[PATH_MAX] = { 0 };
	EFI_SIGNATURE_LIST  *CertList = (EFI_SIGNATURE_LIST *)data;
	EFI_SIGNATURE_DATA  *Cert;
	long count = 0, DataSize = len;
	int size;

	if (dir != NULL) {
		p = &pbuf[0];
		p += snprintf(pbuf, sizeof(pbuf), "%s/", dir);
		pbuf[sizeof(pbuf) - 1] = '\0';
		assert(p < pbuf + sizeof(pbuf));
	}

	certlist_for_each_certentry(CertList, data, size, DataSize) {
		int Index = 0, i, fd = -1;
		const char *ext;

		if (sig != -1 && count != sig)
			continue;


		if (compare_guid(&CertList->SignatureType, &X509_GUID)== 0) {
			ext = "X509";
		} else if (compare_guid(&CertList->SignatureType, &RSA2048_GUID) == 0) {
			ext = "RSA2048";
		} else if (compare_guid(&CertList->SignatureType, &PKCS7_GUID) == 0) {
			ext = "PKCS7";
		} else if (compare_guid(&CertList->SignatureType, &EFI_CERT_SHA256_GUID) == 0) {
			ext = "SHA256";
		} else {
			ext = "Unknown";
		}

		printf("%s: List %ld, type %s\n", name, count++, ext);

		certentry_for_each_cert(Cert, CertList) {
			if (entry != -1 && Index != entry)
				continue;

			if (dir) {
				i = snprintf(p, sizeof(pbuf) - (p - &pbuf[0]),
				    "%s-%d-%ld.%s", name, Index, count, ext);
				pbuf[sizeof(pbuf) - 1] = '\0';
				assert(p + i < &pbuf[0] + sizeof(pbuf));
				printf("saving to %s\n", pbuf);
				fd = open(pbuf, O_WRONLY|O_CREAT|O_TRUNC|O_SYNC,
				    S_IRUSR|S_IRGRP|S_IROTH);
				if (fd == -1) {
					perror("can't open file");
					exit(-1);
				}
			}

			printf("    Signature %d, size %jd, owner %s\n",
			      Index++, CertList->SignatureSize - sizeof(EFI_SIGNATURE_DATA) + 1,
			      guid_to_str(&Cert->SignatureOwner));

			if (strcmp(ext, "X509") == 0) {
				const unsigned char *buf = (unsigned char *)Cert->SignatureData;

				if (fd != -1) {
					ssize_t want, have;

					want = CertList->SignatureSize -
					    sizeof(EFI_SIGNATURE_DATA) + 1;
					have = write(fd, buf, want);
					assert(have == want);
				}

				X509 *X = d2i_X509(NULL, &buf,
						   CertList->SignatureSize - sizeof(EFI_SIGNATURE_DATA) + 1);
				X509_NAME *issuer = X509_get_issuer_name(X);
				X509_NAME *subject = X509_get_subject_name(X);

				printf("        Subject:\n");
				X509_NAME_print_ex_fp(stdout, subject, 12, XN_FLAG_SEP_CPLUS_SPC);
				printf("\n        Issuer:\n");
				X509_NAME_print_ex_fp(stdout, issuer, 12, XN_FLAG_SEP_CPLUS_SPC);
				printf("\n");

			} else if (strcmp(ext, "SHA256") == 0) {
				uint8_t *hash = Cert->SignatureData;
				int j;

				if (fd != -1) {
					ssize_t want, have;

					want = SHA256_DIGEST_SIZE;
					have = write(fd, hash, want);
					assert(have == want);
				}

				printf("        Hash:");
				for (j = 0; j < SHA256_DIGEST_SIZE; j++) {
					printf("%02x", hash[j]);
				}
				printf("\n");
			}

			if (fd != -1) {
				close(fd);
				fd = -1;
			}
		}
	}
}

int
main(int argc, char *argv[])
{
  char *variables[] = { "PK", "KEK", "db", "dbx" , "MokList" };
	char *progname = argv[0], *var = NULL, *file = NULL, *dir = NULL;
	EFI_GUID *owners[] = { &GV_GUID, &GV_GUID, &SIG_DB, &SIG_DB, &MOK_OWNER };
	int i, found = 0, sig = -1, entry = -1, fd;

	while (argc > 1 && argv[1][0] == '-') {
		if (strcmp("--version", argv[1]) == 0) {
			version(progname, NULL);
			exit(0);
		} else if (strcmp("--help", argv[1]) == 0) {
			help(progname);
			exit(0);
		} else if(strcmp(argv[1], "-v") == 0) {
			if (argc <= 2) {
				usage(progname);
				exit(1);
			}
			var = argv[2];
			argv += 2;
			argc -= 2;
		} else if (strcmp(argv[1], "-s") == 0) {
			if (argc <= 2) {
				usage(progname);
				exit(1);
			}
			sscanf(argv[2], "%d-%d", &sig, &entry);
			argv += 2;
			argc -= 2;
		} else if (strcmp(argv[1], "-o") == 0) {
			if (argc <= 2) {
				usage(progname);
				exit(1);
			}
			file = argv[2];
			argv += 2;
			argc -= 2;
		} else if (strcmp(argv[1], "-d") == 0) {
			if (argc <= 2) {
				usage(progname);
				exit(1);
			}
			dir = argv[2];
			argv += 2;
			argc -= 2;
		} else {
			/* unrecognised option */
			break;
		}
	}

	if (argc != 1) {
		usage(progname);
		exit(1);
	}

	if (sig != -1 && !var) {
		fprintf(stderr, "need -v <var> with -s option\n");
		exit(1);
	}

	if (file) {
		fd = open(file, O_CREAT|O_TRUNC|O_WRONLY, 0600);
		if (fd < 0) {
			fprintf(stderr, "failed to open %s: ", file);
			perror("");
			exit(1);
		}
	}

	if (dir) {
		struct stat dstat;
		if (stat(dir, &dstat) != 0) {
			perror("can't stat dir");
			exit(1);
		}
		if (!S_ISDIR(dstat.st_mode)) {
			fprintf(stderr, "%s: is not a dir\n", dir);
			exit(1);
		}
	}

	kernel_variable_init();
	for (i = 0; i < ARRAY_SIZE(owners); i++) {
		int status;
		uint32_t len;
		uint8_t *buf;

		if (var && strcmp(var, variables[i]) != 0)
			continue;

		found = 1;
		status = get_variable_alloc(variables[i], owners[i], NULL,
					    &len, &buf);
		if (status == ENOENT) {
			printf("Variable %s has no entries\n", variables[i]);
			continue;
		} else if (status != 0) {
			printf("Failed to get %s: %d\n", variables[i], status);
			continue;
		}
		printf("Variable %s, length %d\n", variables[i], len);
		if (file)
			write(fd, buf, len);
		else
			parse_db(variables[i], buf, len, sig, entry, dir);
		free(buf);
	}
	if (file)
		close(fd);
	if (!found) {
		fprintf(stderr, "variable %s is not a UEFI secure boot variable\n", var);
		exit(1);
	}
	return 0;
}
		
